@@ -24,7 +24,6 @@ module Control.Monad.Writer.Lazy (
     censor,
     -- * The Writer monad
     Writer,
-    writer,
     runWriter,
     execWriter,
     mapWriter,
@@ -42,7 +41,7 @@ import Control.Monad.Writer.Class
 
 import Control.Monad.Trans
 import Control.Monad.Trans.Writer.Lazy (
-        Writer, writer, runWriter, execWriter, mapWriter,
+        Writer, runWriter, execWriter, mapWriter,
         WriterT(..), execWriterT, mapWriterT)
 
 import Control.Monad
